c -------------------------------------------------------------------- c
c
c   BaySpl:  Baysian Spline Surface Fitting Program For 1-D data
c              Modified-Cholesky decomposition version
c
c   Programed by:  Yasuaki Murata
c                    Geological Survey of Japan, AIST
c                    Geological Information Center
c                    1-1-1 Higashi, Tsukuba,
c                    Ibaraki 305-8567, JAPAN
c                    Email-address:   murata.gsj@aist.go.jp
c
c   History:  Nov, 2002   original coording
c
c   Input control data:
c          Rec#  Format    Attribute
c             1   a50      Title
c             2   a50      point data file name  (read)
c             3   a50      spline coeffient file name  (write)
c             4   free     Xmin, Xmax (calculation window)
c             5   free     Mx, Tx, Np
c                           (knots, normalizing, additional parameter #)
c             6   free     w1, w2, size, epsilon
c                              (initial tradeoff-parameter (w1, w2)
c                               and initial size of triangle for optimi-
c                               zation (size) and convergence epsilon)
c             7   a50     input data format
c
c    Input data file format:
c                           (free format)
c          Col#   Attribute
c             1   data number
c             2   x-value
c             3   y-value
c             4-  additional parameters
c
c    Output spline coeffient file format:
c          Rec#  Format       Attribute
c             1   2i7       Mx, additional parameter #
c             2   6D13.6    sd, w1, w2, Dummy, Dummy, ABIC
c                               (residual(sd), optimized tradeoff param-
c                                eter(w1, w2), final ABIC)
c             4   4f15.7    Xmin, Xmax (calculation window)
c             5-  6D13.6    (S(i), i=1, nSPL), (errS(i), i=1, nSPL)
c                               (spline coefficient and estimation
c                                error)
c            last 2D15.8    Fitting residual.
c
c -------------------------------------------------------------------- c
      program BaySpl1D
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
      parameter (maxH=3, maxH5=maxH+5)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), F(maxN), B(maxN)
      common /BaseI/   nB(maxN, 4)
      common /BaseD/   Bx(maxN, 4)
      common /Limit/   Xmin, Fmin, Xmax, Fmax, Gx
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
      common /ABICcm/  ABIC, w1min, w2min
      common /ABICfn/  SPLdsn, GRDdsn, PNTdsn
      common /Corect/  H(maxN, maxP), Hmin, Hmax
c
      character SPLdsn*70, GRDdsn*70, PNTdsn*70
      dimension hyper(maxH)
      external  funcs
c -------------------------------------------------------------------- c
      ABIC = 9.0d9
      call datred(w1, w2, dist, eps)
      call normal(irc)
      if (irc .ne. 0) stop
      call BxCal
      do 10 i = 1, maxM
        S(i) = 1.0d0
10      continue
      if (eps .eq. 0.0d0) then
        call MABICS(w1, w2, wABIC, 1)
      else
        hyper(1) = dlog(w1)
        hyper(2) = dlog(w2)
        call simplx(hyper, 2, funcs, dist, eps, wABIC)
        call MABICS(w1min, w2min, wABIC, 1)
        end if
c
      stop
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine funcs(nh, hyper, wABIC)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxH=3, maxH5=maxH+5)
      dimension hyper(maxH)
c -------------------------------------------------------------------- c
      w1 = dexp(hyper(1))
      w2 = dexp(hyper(2))
      if (w1 .lt. 1.0d-12) w1 = 1.0d-12
      if (w2 .lt. 1.0d-12) w2 = 1.0d-12
      if (w1 .gt. 1.0d+02) w1 = 1.0d+02
      if (w2 .gt. 1.0d+02) w2 = 1.0d+02
      call MABICS(w1, w2, wABIC, 0)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine datred(w1, w2, dist, eps)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), F(maxN), B(maxN)
      common /Limit/   Xmin, Fmin, Xmax, Fmax, Gx
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
      common /ABICfn/  SPLdsn, GRDdsn, PNTdsn
      common /Corect/  H(maxN, maxP), Hmin, Hmax
c
      dimension Hw(maxP)
      character Title*70, DATdsn*70, Form*70
      character SPLdsn*70, GRDdsn*70, PNTdsn*70
c -------------------------------------------------------------------- c
      write(6, '(a)') ' '
      read(5, '(a70)')  Title
      read(5, '(a70)')  DATdsn
      read(5, '(a70)')  SPLdsn
      read(5, '(a70)')  PNTdsn
      read(5, '(a70)')  GRDdsn
      read(5, *)        Xmin, Xmax, Gx
      read(5, *)        Mx, Tx, nPRM
      read(5, *)        w1, w2, dist, eps
      read(5, '(a70)')  Form
c
      do i = 69, 1, -1
        if (Title(i:i+1)  .eq. '/*') Title(i:70) = ''
        if (DATdsn(i:i+1) .eq. '/*') DATdsn(i:70) = ''
        if (SPLdsn(i:i+1) .eq. '/*') SPLdsn(i:70) = ''
        if (PNTdsn(i:i+1) .eq. '/*') PNTdsn(i:70) = ''
        if (GRDdsn(i:i+1) .eq. '/*') GRDdsn(i:70) = ''
        end do
c
      nSPL = Mx + 3
      Dx = Tx / dfloat(Mx)
c
      write(6, '(14h Title    :   , a50)') Title
      write(6, '(14h Data   file: , a50)') DATdsn
      write(6, '(14h Spline file: , a50)') SPLdsn
      write(6, '(14h Window:      , 2f13.2)') Xmin, Xmax
      write(6, '(14h Spline:      , i6, f10.5, i5)') Mx, Tx, nPRM
      write(6, '(14h Hyper-parm:  , 4D15.8)') w1, w2, dist, eps
c
      if (nSPL+nPRM .gt. maxM) then
        write(6, *) ' Total spline knots Mx+3+nPRM > maxM ',
     &    Mx+3+nPRM
        stop
        end if
      if (nPRM .gt. maxP) then
        write(6, *) 'nPRM > maxP'
        stop
        end if
c
      im = 0
      if (Xmin .eq. Xmax) im = 1
c
      open(12, file=DATdsn)
      rewind(12)
c
      if (im .eq. 1) then
        Xmin =  9.0d9
        Xmax = -9.0d9
        end if
      Fmin =  9.0d9
      Fmax = -9.0d9
c
      nDAT = 0
c
2     read(12, Form, end=1) Nw, Xw, Fw, (Hw(i), i=1, nPRM)
      if (Nw .lt. 0) go to 2
      if (Nw .eq. 0) go to 3
      if (im .eq. 1) then
        if (Xmin .gt. Xw) Xmin = Xw - 0.1d0
        if (Xmax .lt. Xw) Xmax = Xw + 0.1d0
      else
        if (Xw .lt. Xmin .or. Xw .gt. Xmax) go to 2
        end if
      if (Fmin .gt. Fw) Fmin = Fw
      if (Fmax .lt. Fw) Fmax = Fw
c
3     nDAT = nDAT + 1
      if (nDAT .le. maxN) then
        N(nDAT) = Nw
        X(nDAT) = Xw
        F(nDAT) = Fw
        do 10 i = 1, nPRM
          H(nDAT, i) = Hw(i)
10        continue
      else
        write(6, *) ' #### nDAT > maxN stop #### ', maxN, nDAT
        stop
        end if
      go to 2
1     continue
      close(12)
c
      write(6, '(a)') ' '
      write(6, '(a20, i7)')
     &         ' nDAT    =             ', nDAT
      write(6, '(a20, 3f12.2)')
     &         ' Xmin, Xmax, Xwid = ', Xmin, Xmax, Xmax-Xmin
      write(6, '(a20, 3f12.2)')
     &         ' Fmin, Fmax, Fwid = ', Fmin, Fmax, Fmax-Fmin
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine MABICS(w1, w2, wABIC, ip)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common /DataI/   N(maxN), nDAT
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
      common /ABICcm/  ABIC, w1min, w2min
      common /ABICfn/  SPLdsn, GRDdsn, PNTdsn
c
      character  SPLdsn*70, GRDdsn*70, PNTdsn*70
      data       pi/3.1415926535898d0/
c -------------------------------------------------------------------- c
      write(6, '(a)') ' '
      write(6, *) ' MABICS start '
      call PRMcal(w1, w2, detZ, ip)
      if (ip .eq. 1 .and. PNTdsn .ne. 'dummy') then
        open(13, file=PNTdsn)
        call ERRcal(w1, w2, rss, rsr, detD, ip)
        close(13)
      else
        call ERRcal(w1, w2, rss, rsr, detD, 0)
        end if
      sigma2 = (rss + rsr) / dfloat(nDAT)
      sd = dsqrt(sigma2)
      wABIC = dfloat(nDAT) * dlog(2.0d0 * pi * sigma2)
     &                 - detD + detZ + dfloat(nDAT)
      write(6, '(a25, 3D15.7, 2f12.3)')
     &  ' sgm2, detD, detZ, ABIC=', sigma2, detD, detZ, wABIC
c
      do 20 i = 1, nSPL + nPRM
        errS(i) = errS(i) * sd
20      continue
c
      if (wABIC .lt. ABIC) then
        w1min  = w1
        w2min  = w2
        ABIC   = wABIC
        do 30 i = 1, nPRM
          write(6, '(a, i5, f10.3, 4h +- , f10.3)')
     &       ' Prm = ', i, S(nSPL+i), errS(nSPL+i)
30        continue
        end if
c
      if (ip .eq. 1 .and. SPLdsn .ne. 'dummy') then
        open(10, file=SPLdsn)
        rewind(10)
        call SPLwtr(10, sd, rss, w1, w2, wABIC)
        close(10)
        write(6, '(a)') ' '
        end if
c
      if (ip .eq. 1 .and. GRDdsn .ne. 'dummy') then
        open(10, file=GRDdsn)
        rewind(10)
        call GRDwtr(10, sd, rss, w1, w2, wABIC)
        close(10)
        write(6, '(a)') ' '
        end if
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine PRMcal(w1, w2, detZ, ip)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
c
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
c -------------------------------------------------------------------- c
      write(6, '(a, 2D25.18, a)')
     &  '  PRMcal start (w1, w2 = ', w1, w2, ')'
c
      call Zset(w1, w2)
      call solveZ(S, nSPL, nPRM, 1)
      call lndetU(nSPL, detZ)
c
c     if (ip .eq. 1 .and. SPLdsn .ne. 'dummy')
c    &       call errorS(errS, nSPL, nPRM)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine Zset(w1, w2)
c
      implicit   real*8 (a-h, o-z)
c -------------------------------------------------------------------- c
      write(6, *) ' Zset start '
c
      call clearZ
      call obserZ
      call priorZ(w1, w2, dum, 0)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine ERRcal(w1, w2, rss, rsr, detD, ip)
c
      implicit real*8 (a-h, o-z)
c -------------------------------------------------------------------- c
      write(6, *) ' ERRcal start '
c
      call SPLerr(rss, ip)
      call SMTerr(w1, w2, rsr, detD)
      write(6, '(a, 2D15.7)') ' Fitness, Smoothness    =', rss, rsr
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine SPLerr(rss, ip)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), F(maxN), B(maxN)
      common /BaseI/   nB(maxN, 4)
      common /BaseD/   Bx(maxN, 4)
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
      common /Corect/  H(maxN, maxP), Hmin, Hmax
c -------------------------------------------------------------------- c
      write(6, *) ' SPLerr start '
c
      rss = 0.0d0
      do 10 k = 1, nDAT
        if (N(k) .le. 0) go to 10
        totalS = 0.0d0
        do 20 i = 1, 4
          totalS = totalS + S(nB(k, i)) * Bx(k, 5-i)
20        continue
        B(k)  = F(k)
        do 30 i = 1, nPRM
          B(k)  = B(k) - H(k, i) * S(nSPL+i)
30        continue
        if (ip .eq. 1) then
          Xw = X(k) * (Xmax - Xmin) + Xmin
          write(13, '(i5, f12.2, 3f9.3)')
     &      N(k), Xw, totalS, B(k), B(k) - totalS
          end if
        rs  = B(k) - totalS
        rss = rss + rs**2
10      continue
      sdx = dsqrt(rss / dfloat(nDAT))
      write(6, '(a, f12.7)') '  Fitting standard deviation = ', sdx
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine SPLwtr(ifile, sd, rss, w1, w2, wABIC)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
c
      common /DataI/   N(maxN), nDAT
      common /Limit/   Xmin, Fmin, Xmax, Fmax, Gx
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
c -------------------------------------------------------------------- c
      write(6, *) ' SPLwtr start '
c
      write(ifile, '(2i7)') Mx, nPRM
      write(ifile, '(5D13.6, f13.4)') sd, w1, w2, 0.0d0, 0.0d0, wABIC
      write(ifile, '(2f15.2)') Xmin, Xmax
      write(ifile, '(6D13.6)') (S(i), i=1, nSPL), (errS(i), i=1, nSPL)
      write(ifile, '(a20, 2D15.8)') '       rss,  sd =   ',
     &            rss, dsqrt(rss / dfloat(nDAT))
      write(ifile, '(5h PRM ,i5, 2D13.6)')
     &       (i, S(i+nSPL), errS(i+nSPL), i=1, nPRM)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine GRDwtr(ifile, sd, rss, w1, w2, wABIC)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
c
      common /DataI/   N(maxN), nDAT
      common /Limit/   Xmin, Fmin, Xmax, Fmax, Gx
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
c
      Dimension Bx2(4)
c -------------------------------------------------------------------- c
      write(6, *) ' GRDwtr start '
c
      Nx = (Xmax - Xmin) / Gx + 1
c
      do 20 ix = 1, Nx
        xw = dfloat(ix-1) * Gx / (Xmax - Xmin)
        if (xw .gt. 1.0d0) go to 20
        rx = xw / Dx
        lx = int(rx) + 1
        if (xw .eq. 1.0d0) lx = Mx
        rx = rx - dfloat(lx - 1)
        call Bsp3(rx, Bx2)
        g = 0.0d0
        do 40 jx = 1, 4
          g = g + S(lx+jx-1) * Bx2(5-jx)
40        continue
        write(ifile, 100) ix, dfloat(ix-1)*Gx+Xmin, g
100     format(i5, f12.2, f9.3)
20      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine BxCal
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), F(maxN), B(maxN)
      common /BaseI/   nB(maxN, 4)
      common /BaseD/   Bx(maxN, 4)
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
c
      dimension Bx2(4)
c -------------------------------------------------------------------- c
      do 10 i = 1, nDAT
        if (N(i) .le. 0) go to 10
        rx = X(i) / Dx
        nx = int(rx) + 1
        if (X(i) .ge. 1.0d0) nx = Mx
        rx = rx - dfloat(nx - 1)
        if (rx.lt.0.0d0 .or. rx.gt.1.0d0) write(6, *) 'BxCal Error'
        do 20 ix = 1, 4
          nB(i, ix) = nx + ix - 1
20        continue
        call Bsp3(rx, Bx2)
        do 30 j = 1, 4
          Bx(i, j) = Bx2(j)
30        continue
10      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine normal(irc)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), F(maxN), B(maxN)
      common /Limit/   Xmin, Fmin, Xmax, Fmax, Gx
c -------------------------------------------------------------------- c
      irc = 0
      if (nDAT .lt. 20) then
        irc = 999
        write(6, *) '#### Number of points < 20 ####', nDAT
        return
        end if
      do 10 i = 1, nDAT
        if (N(i) .gt. 0) then
          X(i) = (X(i) - Xmin) / (Xmax - Xmin)
          end if
10      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
c   Modified Cholesky compact version   U(i,i) = 1                     c
c              use U2 dimenstion                                       c
c -------------------------------------------------------------------- c
      subroutine solveZ(S, nSPL, nPRM, iback)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common /matrix/ U(nRcmp, nCcmp), nROW
c
      dimension S(maxM), w(maxM)
c -------------------------------------------------------------------- c
      write(6, *) ' solveZ start (Compact modified Cholesky)'
c
      m = nSPL + nPRM
      do 10 k = 1, nSPL
        nstr = max(   1, k-nROW+1)
        nfin = min(nSPL, k+nROW-1)
        rU = U(1, k)
        do 20 i = nstr, k-1
          kri = k - i + 1
          w(i) = U(kri, i) * U(1, i)
          rU = rU - w(i) * U(kri, i)
20        continue
        U(1, k) = rU
c       if (rU .lt. 0.0) U(1, k) = -rU
        do 30 j = k+1, nfin
          jrk = j - k + 1
          rU = U(jrk, k)
          do 40 i = max(1, j-nROW+1), k-1
            rU = rU - w(i) * U(j-i+1, i)
40          continue
          U(jrk, k) = rU / U(1, k)
30        continue
        if (nPRM .ne. 0) then
          do 31 j = nSPL+1, m
            jr = nROW + j - nSPL
            rU = U(jr, k)
            do 41 i = nstr, k-1
              rU = rU - w(i) * U(jr, i)
41            continue
            U(jr, k) = rU / U(1, k)
31          continue
          end if
10      continue
c
      if (nPRM .ne. 0) then
        do 12 k = nSPL+1, m
          kr = nROW + k - nSPL
          rU = U(kr, k)
          do 22 i = 1, k-1
            w(i) = U(kr, i) * U(1, i)
            rU = rU - w(i) * U(kr, i)
22          continue
          U(1, k) = rU
          do 32 j = k+1, m
            jr = nROW + j - nSPL
            rU = U(jr, k)
            do 42 i = 1, k-1
              rU = rU - w(i) * U(jr, i)
42            continue
            U(jr, k) = rU / U(1, k)
32          continue
12        continue
        end if
c
      if (iback .eq. 1) then
        do 50 k = 1, nSPL
          rU = U(nROW+nPRM+1, k)
          do 60 i = max(1, k-nROW+1), k-1
            rU = rU - U(k-i+1, i) * S(i)
60          continue
          S(k) = rU
50        continue
        if (nPRM .ne. 0) then
          do 51 k = nSPL+1, m
            kr = nROW + k - nSPL
            rU = U(nROW+nPRM+1, k)
            do 61 i = 1, k-1
              rU = rU - U(kr, i) * S(i)
61            continue
            S(k) = rU
51          continue
          end if
        if (nPRM .ne. 0) then
          do 71 k = m, nSPL+1, -1
            rU = S(k) / U(1, k)
            do 81 i = k+1, m
              rU = rU - U(nROW+i-nSPL, k) * S(i)
81            continue
            S(k) = rU
71          continue
          end if
        do 70 k = nSPL, 1, -1
          rU = S(k) / U(1, k)
          do 80 i = k+1, min(nSPL, nROW+k-1)
            rU = rU - U(i-k+1, k) * S(i)
80          continue
          if (nPRM .ne. 0) then
            do 82 i = nSPL+1, m
              rU = rU - U(nROW+i-nSPL, k) * S(i)
82            continue
            end if
          S(k) = rU
70        continue
        end if
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine lndetU(m, detU)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common /matrix/ U(nRcmp, nCcmp), nROW
c -------------------------------------------------------------------- c
      write(6, *) ' lndetU start '
c
      detU = 0.0d0
      do 10 nC = 1, m
        if (U(1, nC) .gt. 0.0d0) then
          detU = detU + dlog(U(1, nC))
        else
          detU = detU + dlog(-U(1, nC))
          write(6, *) ' lndetU U(1, nC) < 0.0', nC , m, U(1, nC)
          end if
10      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine retrnE(ePRM, nC)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common /matrix/ U(nRcmp, nCcmp), nROW
c -------------------------------------------------------------------- c
      write(6, *) ' retrnE start '
c
      ePRM = dsqrt(1.0d0 / U(1, nC))
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine errorS(errS, nSPL, nPRM)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common  /matrix/ U(nRcmp, nCcmp), nROW
c
      dimension  errS(maxM), Z(maxM)
c -------------------------------------------------------------------- c
      write(6, *) ' errorS start '
c
      m = nSPL + nPRM
      do 10 k = 1, nSPL
        Z(k) = 1.0d0
        do 20 j = k+1, nSPL
          rU = 0.0d0
          do 30 i = max(k, j-nROW+1), j-1
            rU = rU - U(j-i+1, i) * Z(i)
30          continue
          Z(j) = rU
20        continue
        if (nPRM .ne. 0) then
          do 21 j = nSPL+1, m
            jr = nROW + j - nSPL
            rU = 0.0d0
            do 31 i = k, j-1
              rU = rU - U(jr, i) * Z(i)
31            continue
            Z(j) = rU
21          continue
          end if
        if (nPRM .ne. 0) then
          do 41 j = m, nSPL+1, -1
            rU = Z(j) / U(1, j)
            do 51 i = j+1, m
              ir = nROW + i - nSPL
              rU = rU - U(ir, j) * Z(i)
51            continue
            Z(j) = rU
41          continue
          end if
        do 40 j = nSPL, k, -1
          rU = Z(j) / U(1, j)
          do 50 i = j+1, min(nSPL, j+nROW-1)
            rU = rU - U(i-j+1, j) * Z(i)
50          continue
          if (nPRM .ne. 0) then
            do 52 i = nSPL+1, m
              ir = nROW + i - nSPL
              rU = rU - U(ir, j) * Z(i)
52            continue
            end if
          Z(j) = rU
40        continue
        errS(k) = Z(k)
10      continue
c
      do 13 k = nSPL+1, m
        Z(k) = 1.0d0
        do 23 j = k+1, m
          jr = nROW + j - nSPL
          rU = 0.0d0
          do 33 i = k, j-1
            rU = rU - U(jr, i) * Z(i)
33          continue
          Z(j) = rU
23        continue
        do 43 j = m, k, -1
          rU = Z(j) / U(1, j)
          do 53 i = j+1, m
            ir = nROW + i - nSPL
            rU = rU - U(ir, j) * Z(i)
53          continue
          Z(j) = rU
43        continue
        errS(k) = Z(k)
13      continue
      do 60 k = 1, m
        errS(k) = dsqrt(dabs(errS(k)))
60      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine obserZ
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), F(maxN), B(maxN)
      common /BaseI/   nB(maxN, 4)
      common /BaseD/   Bx(maxN, 4)
      common /SplineI/ nSPL, nPRM, Mx
      common /Corect/  H(maxN, maxP), Hmin, Hmax
      common /matrix/  U(nRcmp, nCcmp), nROW
      common /matrix2/ U2(nRcmp, nCcmp), nU2
c -------------------------------------------------------------------- c
      write(6, *) ' obserZ start '
c
      nROW = 4
      if (nU2 .eq. 999) go to 1
c
      do 10 i = 1, nDAT
        do 20 j = 1, nPRM
          k = nSPL + j
          U(nROW+nPRM+1, k) = U(nROW+nPRM+1, k) + H(i, j) * F(i)
          do 20 l = j, nPRM
            m = nROW + l
            U(m, k) = U(m, k) + H(i, j) * H(i, l)
20        continue
        if (N(i) .le. 0) go to 10
        do 30 ix1 = 1, 4
          n1 = nB(i, ix1)
          p1 = Bx(i, 5-ix1)
          do 40 j = 1, nPRM
            U(nROW+j, n1) = U(nROW+j, n1) + p1 * H(i, j)
40          continue
          U(nROW+nPRM+1, n1) = U(nROW+nPRM+1, n1) + p1 * F(i)
          do 50 ix2 = ix1, 4
            n2 = nB(i, ix2)
            n3 = n2 - n1 + 1
            p2 = Bx(i, 5-ix2)
            U(n3, n1) = U(n3, n1) + p1 * p2
50          continue
30        continue
10      continue
c
      do 60 i = 1, nCcmp
        do 60 j = 1, nRcmp
          U2(j, i) = U(j, i)
60      continue
      nU2 = 999
      return
c
1     continue
      do 70 iy1 = 1, nCcmp
        do 70 ix1 = 1, nRcmp
          U(ix1, iy1) = U2(ix1, iy1)
70      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine clearZ
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common /matrix/ U(nRcmp, nCcmp), nROW
c -------------------------------------------------------------------- c
      write(6, *) ' clearZ start '
c
      do 10 nC = 1, nCcmp
        do 10 nR = 1, nRcmp
          U(nR, nC) = 0.0d0
10      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine SMTerr(w1, w2, rsr, detD)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
      common /matrix/  U(nRcmp, nCcmp), nROW
c -------------------------------------------------------------------- c
      write(6, *) ' SMTerr start '
c
      call clearZ
      call priorZ(w1, w2, rsr, 1)
      call solveZ(S, nSPL, 0, 0)
      call lndetU(nSPL-1, detD)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine priorZ(w1, w2, rsr, ip)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=100000, maxMx=106000, maxP=10)
      parameter (maxM=maxMx+3+maxP)
      parameter (nRcmp=5+maxP, nCcmp=maxM+1)
c
      common /SplineI/ nSPL, nPRM, Mx
      common /SplineD/ S(maxM), errS(maxM), Dx
      common /matrix/  U(nRcmp, nCcmp), nROW
c
      dimension bi(4, 4)
c -------------------------------------------------------------------- c
      write(6, '(a, 2d15.6)') '  priorZ start ', w1, w2
c
      call intB(bi, Dx, w1, w2)
c
      if (ip .eq. 1) rsr = 0.0d0
c
      do 10 i = 1, Mx
        do 10 k = 0, 3
          n1 = i + k
          do 10 l = k, 3
            n2 = i + l
            w = bi(4-k, 4-l)
            n3 = n2 - n1 + 1
            U(n3, n1) = U(n3, n1) + w
            if (ip .eq. 1) then
              if (n1 .ne. n2) w = w * 2.0d0
              rsr = rsr + S(n1) * S(n2) * w
              end if
10      continue
      if (ip .eq. 1 .and. rsr .lt. 0.0d0) then
        write(6, *) 'rsr < 0.0', rsr
c       rsr = 0.0d0
        rsr = -rsr
        end if
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine Bsp3(r, B)
c
      implicit  real*8 (a-h, o-z)
      dimension B(4)
c -------------------------------------------------------------------- c
      r2 = r * r
      r3 = r * r * r
c
      a = 1.0d0 / 6.0d0
      B(1) = r3 * a
      B(2) = (3.0d0 * (-r3 + r2 + r) + 1.0d0) * a
      B(3) = (3.0d0 * r3 - 6.0d0 * r2 + 4.0d0) * a
      B(4) = (-r3 + 3.0d0 * (r2 - r) + 1.0d0) * a
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine intB(bi, Dx, w1, w2)
c
      implicit  real*8 (a-h, o-z)
c
      dimension bi(4,4), a(3,4), b(3,4), c(3,4), d(3,4), fac(3)
      data a   / 1.0d0, 0.0d0, 0.0d0,-3.0d0, 0.0d0, 0.0d0,
     &           3.0d0, 0.0d0, 0.0d0,-1.0d0, 0.0d0, 0.0d0/
      data b   / 0.0d0, 1.0d0, 0.0d0, 3.0d0,-3.0d0, 0.0d0,
     &          -6.0d0, 3.0d0, 0.0d0, 3.0d0,-1.0d0, 0.0d0/
      data c   / 0.0d0, 0.0d0, 1.0d0, 3.0d0, 2.0d0,-3.0d0,
     &           0.0d0,-4.0d0, 3.0d0,-3.0d0, 2.0d0,-1.0d0/
      data d   / 0.0d0, 0.0d0, 0.0d0, 1.0d0, 1.0d0, 1.0d0,
     &           4.0d0, 0.0d0,-2.0d0, 1.0d0,-1.0d0, 1.0d0/
      data fac /36.0d0, 4.0d0, 1.0d0/
c -------------------------------------------------------------------- c
      do 10 k = 1, 4
      do 10 l = 1, 4
c
        r1 =
     1   (a(2,k)*a(2,l)                                          )/7.0d0
     2  +(a(2,k)*b(2,l)+b(2,k)*a(2,l)                            )/6.0d0
     3  +(a(2,k)*c(2,l)+b(2,k)*b(2,l)+c(2,k)*a(2,l)              )/5.0d0
     4  +(a(2,k)*d(2,l)+b(2,k)*c(2,l)+c(2,k)*b(2,l)+d(2,k)*a(2,l))/4.0d0
     5  +(              b(2,k)*d(2,l)+c(2,k)*c(2,l)+d(2,k)*b(2,l))/3.0d0
     6  +(                            c(2,k)*d(2,l)+d(2,k)*c(2,l))/2.0d0
     7  +(                                          d(2,k)*d(2,l))/1.0d0
        r2 =
     1   (a(3,k)*a(3,l)                                          )/7.0d0
     2  +(a(3,k)*b(3,l)+b(3,k)*a(3,l)                            )/6.0d0
     3  +(a(3,k)*c(3,l)+b(3,k)*b(3,l)+c(3,k)*a(3,l)              )/5.0d0
     4  +(a(3,k)*d(3,l)+b(3,k)*c(3,l)+c(3,k)*b(3,l)+d(3,k)*a(3,l))/4.0d0
     5  +(              b(3,k)*d(3,l)+c(3,k)*c(3,l)+d(3,k)*b(3,l))/3.0d0
     6  +(                            c(3,k)*d(3,l)+d(3,k)*c(3,l))/2.0d0
     7  +(                                          d(3,k)*d(3,l))/1.0d0
c
        bi(k,l) = r1 / fac(2) * w1 / Dx + r2 / fac(3) * w2 / Dx**3
10    continue
c
      return
      end
