c -------------------------------------------------------------------- c
c
c   BaySpl:  Baysian Spline Surface Fitting Program For 2-D data
c              Modified-Cholesky decomposition version
c
c   Programed by:  Yasuaki Murata
c                    GEOLOGICAL SURVEY OF JAPAN
c                    Geological Information Center
c                    1-1-1 Higashi, Tsukuba,
c                    Ibaraki 305-8567, JAPAN
c                    Email-address:   murata.gsj@aist.go.jp
c
c   History:  May, 1990   original coording
c             January, 1992  modified-Cholesky version
c             February, 1992  change input data format
c
c   Input control data:
c          Rec#  Format       Attribute
c             1   a70      Title
c             2   a70      gravity data file name  (read)
c             3   a70      spline coeffient file name  (write)
c             4   a70      point file name  (write)
c             5   a70      grid file name  (write)
c             6   free     Xmin, Ymin, Xmax, Ymax, Dx, Dy (calculation window)
c             7   free     Mx, My, Tx, Ty, Np
c                              (# of knots and normalized coordinates)
c             8   free     w1, w2, size, epsilon
c                              (initial tradeoff-parameter (w1, w2)
c                               and initial size of triangle for optimi-
c                               zation (size) and convergence epsilon)
c             9   text     data read format
c
c    Input data file format:
c                           (free format)
c          Col#      Attribute
c             1   data number
c             2   x-location
c             3   y-location
c             4   z-value
c             5-  additional parameters
c
c    Output spline coeffient file format:
c          Rec#  Format       Attribute
c             1   3i5       Mx, My, # of additional parameters
c             2   6D13.6    sd, w1, w2, Dummy, Dummy, ABIC
c                               (residual(sd), optimized tradeoff param-
c                                eter(w1, w2), final ABIC)
c             4   4f15.7    Xmin, Ymin, Xmax, Ymax (calculation window)
c             5-  6D13.6    (S(i), i=1, nSPL), (errS(i), i=1, nSPL)
c                               (spline coefficient and estimation
c                                error)
c             lst 2D15.8    Fitting residual.
c
c -------------------------------------------------------------------- c
      program BaySpl2D
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
      parameter (maxH=3, maxH5=maxH+5)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), Y(maxN), F(maxN), B(maxN)
      common /BaseI/   nB(maxN, 4, 4)
      common /BaseD/   Bx(maxN, 4), By(maxN, 4)
      common /Limit/   Xmin, Ymin, Fmin, Xmax, Ymax, Fmax, Gx, Gy
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
      common /ABICcm/  ABIC, w1min, w2min
      common /ABICfn/  SPLdsn, GRDdsn, PNTdsn, Form
      common /Corect/  H(maxN, maxP)
c
      character SPLdsn*70, GRDdsn*70, PNTdsn*70, Form*70
      dimension hyper(maxH)
      external  funcs
c -------------------------------------------------------------------- c
      ABIC = 9.0d9
      call datred(w1, w2, dist, eps)
      call normal(irc)
      if (irc .ne. 0) stop
      call BxBy
      do 10 i = 1, maxM
        S(i) = 1.0d0
10      continue
      if (eps .eq. 0.0d0) then
        call MABICS(w1, w2, wABIC, 1)
      else
        hyper(1) = dlog(w1)
        hyper(2) = dlog(w2)
        call simplx(hyper, 2, funcs, dist, eps, wABIC)
        call MABICS(w1min, w2min, wABIC, 1)
        end if
c
      stop
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine funcs(nh, hyper, wABIC)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxH=3, maxH5=maxH+5)
      dimension hyper(maxH)
c -------------------------------------------------------------------- c
      w1 = dexp(hyper(1))
      w2 = dexp(hyper(2))
      if (w1 .lt. 1.0D-10) w1 = 1.0D-10
      if (w2 .lt. 1.0D-10) w2 = 1.0D-10
      if (w1 .gt. 1.0d+02) w1 = 1.0d+02
      if (w2 .gt. 1.0d+02) w2 = 1.0d+02
      call MABICS(w1, w2, wABIC, 0)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine datred(w1, w2, dist, eps)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), Y(maxN), F(maxN), B(maxN)
      common /Limit/   Xmin, Ymin, Fmin, Xmax, Ymax, Fmax, Gx, Gy
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
      common /ABICfn/  SPLdsn, GRDdsn, PNTdsn, Form
      common /Corect/  H(maxN, maxP)
c
      dimension Hw(maxP), Hmin(maxP), Hmax(maxP)
      character Title*70, DATdsn*70
      character SPLdsn*70, GRDdsn*70, PNTdsn*70, Form*70
c -------------------------------------------------------------------- c
      write(6, '(a)') ' '
      read(5, '(a70)')  Title
      read(5, '(a70)')  DATdsn
      read(5, '(a70)')  SPLdsn
      read(5, '(a70)')  PNTdsn
      read(5, '(a70)')  GRDdsn
      read(5, *)        Xmin, Ymin, Xmax, Ymax, Gx, Gy
      read(5, *)        Mx, My, Tx, Ty, nPRM
      read(5, *)        w1, w2, dist, eps
      read(5, '(a70)')  Form
c
      do i = 69, 1, -1
        if (Title(i:i+1)  .eq. '/*') Title(i:70) = ''
        if (DATdsn(i:i+1) .eq. '/*') DATdsn(i:70) = ''
        if (SPLdsn(i:i+1) .eq. '/*') SPLdsn(i:70) = ''
        if (PNTdsn(i:i+1) .eq. '/*') PNTdsn(i:70) = ''
        if (GRDdsn(i:i+1) .eq. '/*') GRDdsn(i:70) = ''
        end do
c
      if (Mx .gt. maxMx) then
        write(0, *) 'Mx .gt. maxMx'
        stop
        end if
      if (My .gt. maxMy) then
        write(0, *) 'My .gt. maxMy'
        stop
        end if
      if (nPRM .gt. maxP) then
        write(0, *) 'nPRM .gt. maxP'
        stop
        end if
c
      do i = 1, nPRM
        Hmin(i) =  9.0d29
        Hmax(i) = -9.0d29
        end do
c
      nSPL = (Mx + 3) * (My + 3)
      Dx = Tx / dfloat(Mx)
      Dy = Ty / dfloat(My)
c
      write(6, '(14h Title    :   , a50)') Title
      write(6, '(14h Data   file: , a50)') DATdsn
      write(6, '(14h Spline file: , a50)') SPLdsn
      write(6, '(14h Window:      , 4f13.2)') Xmin, Ymin, Xmax, Ymax
      write(6, '(14h Spline:      , 2i6, 2f10.5, i5)') Mx,My,Tx,Ty,nPRM
      write(6, '(14h Hyper-parm:  , 4D15.8)') w1, w2, dist, eps
c
      if (nSPL+nPRM .gt. maxM) then
        write(6, *) ' Total spline knots (Mx+3)*(My+3)+nPRM > maxM ',
     &    (Mx+3)*(My+3)+nPRM
        stop
        end if
      if (nPRM .gt. maxP) then
        write(6, *) 'nPRM > maxP'
        stop
        end if
c
      im = 0
      if (Xmin .eq. Xmax) im = 1
c
      open(12, file=DATdsn)
      rewind(12)
c
      if (im .eq. 1) then
        Xmin =  9.0d9
        Ymin =  9.0d9
        Xmax = -9.0d9
        Ymax = -9.0d9
        end if
      Fmin =  9.0d9
      Fmax = -9.0d9
c
      nDAT = 0
c
2     read(12, Form, end=1) Nw, Xw, Yw, Fw, (Hw(i), i=1, nPRM)
      if (Nw .lt. 0) go to 2
      if (Nw .eq. 0) go to 3
      if (im .eq. 1) then
        if (Xmin .gt. Xw) Xmin = Xw - 0.1D0
        if (Ymin .gt. Yw) Ymin = Yw - 0.1D0
        if (Xmax .lt. Xw) Xmax = Xw + 0.1D0
        if (Ymax .lt. Yw) Ymax = Yw + 0.1D0
      else
        if (Xw .lt. Xmin .or. Xw .gt. Xmax) go to 2
        if (Yw .lt. Ymin .or. Yw .gt. Ymax) go to 2
        end if
      if (Fmin .gt. Fw) Fmin = Fw
      if (Fmax .lt. Fw) Fmax = Fw
c
3     nDAT = nDAT + 1
      if (nDAT .le. maxN) then
        N(nDAT) = Nw
        X(nDAT) = Xw
        Y(nDAT) = Yw
        F(nDAT) = Fw
        do 20 i = 1, nPRM
          H(nDAT, i) = Hw(i)
          if (Hmin(i) .gt. Hw(i)) Hmin(i) = Hw(i)
          if (Hmax(i) .lt. Hw(i)) Hmax(i) = Hw(i)
20        continue
      else
        write(6, *) ' #### nDAT > maxN stop #### ', maxN, nDAT
        stop
        end if
      go to 2
1     continue
      close(12)
c
      write(6, '(a)') ' '
      write(6, '(a20, i7)')
     &         ' nDAT    =             ', nDAT
      write(6, '(a20, 3f12.2)')
     &         ' Xmin, Xmax, Xwid = ', Xmin, Xmax, Xmax-Xmin
      write(6, '(a20, 3f12.2)')
     &         ' Ymin, Ymax, Ywid = ', Ymin, Ymax, Ymax-Ymin
      write(6, '(a20, 3f12.2)')
     &         ' Fmin, Fmax, Fwid = ', Fmin, Fmax, Fmax-Fmin
      do i = 1, nPRM
        write(6, '(a20, i5, 3f12.6)')
     &   ' Hmin, Hmax, Hwid = ', i, Hmin(i), Hmax(i), Hmax(i)-Hmin(i)
        end do
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine MABICS(w1, w2, wABIC, ip)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common /DataI/   N(maxN), nDAT
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
      common /ABICcm/  ABIC, w1min, w2min
      common /ABICfn/  SPLdsn, GRDdsn, PNTdsn, Form
c
      character  SPLdsn*70, GRDdsn*70, PNTdsn*70, Form*70
      data       pi/3.1415926535898d0/
c -------------------------------------------------------------------- c
      write(6, '(a)') ' '
      write(6, *) ' MABICS start '
      call PRMcal(w1, w2, detZ, ip)
      if (ip .eq. 1 .and. PNTdsn .ne. 'dummy') then
        open(13, file=PNTdsn)
        call ERRcal(w1, w2, rss, rsr, detD, ip)
        close(13)
      else
        call ERRcal(w1, w2, rss, rsr, detD, 0)
        end if
      sigma2 = (rss + rsr) / dfloat(nDAT)
      sd = dsqrt(sigma2)
      wABIC = dfloat(nDAT) * dlog(2.0d0 * pi * sigma2)
     &                 - detD + detZ + dfloat(nDAT)
      write(6, '(a25, 3D15.7, 2f10.3)')
     &  ' sgm2, detD, detZ, ABIC =', sigma2, detD, detZ, wABIC
c
      do 20 i = 1, nSPL + nPRM
        errS(i) = errS(i) * sd
20      continue
c
      if (wABIC .lt. ABIC) then
        w1min  = w1
        w2min  = w2
        ABIC   = wABIC
        do 30 i = 1, nPRM
          write(6, '(a, i5, f10.3, 4h +- , f10.3)')
     &       ' Prm = ', i, S(nSPL+i), errS(nSPL+i)
30        continue
        end if
c
      if (ip .eq. 1 .and. SPLdsn .ne. 'dummy') then
        open(10, file=SPLdsn)
        rewind(10)
        call SPLwtr(10, sd, rss, w1, w2, wABIC)
        close(10)
        end if
c
      if (ip .eq. 1 .and. GRDdsn .ne. 'dummy') then
        open(11, file=GRDdsn)
        rewind(11)
        call GRDwtr(11, sd, rss, w1, w2, wABIC)
        close(11)
        end if
c
      write(6, '(a)') ' '
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine PRMcal(w1, w2, detZ, ip)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
c
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
      common /ABICfn/  SPLdsn, GRDdsn, PNTdsn, Form
      character SPLdsn*70, GRDdsn*70, PNTdsn*70, Form*70
c -------------------------------------------------------------------- c
      write(6, '(a, 2D20.13, a)')
     &  ' PRMcal start ( w1, w2 = ', w1, w2, ' )'
c
      call Zset(w1, w2)
      call solveZ(S, nSPL, nPRM, 1)
      call lndetU(nSPL, detZ)
c
c Estimation Error Calculation
c
c     if (ip .eq. 1 .and. SPLdsn .ne. 'dummy')
c    &       call errorS(errS, nSPL, nPRM)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine Zset(w1, w2)
c
      implicit    real*8 (a-h, o-z)
c -------------------------------------------------------------------- c
      write(6, *) ' Zset start '
c
      call clearZ
      call obserZ
      call priorZ(w1, w2, dum, 0)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine ERRcal(w1, w2, rss, rsr, detD, ip)
c
      implicit real*8 (a-h, o-z)
c -------------------------------------------------------------------- c
      write(6, *) ' ERRcal start '
c
      call SPLerr(rss, ip)
      call SMTerr(w1, w2, rsr, detD)
      write(6, '(a, 2D15.7)') ' Fitness, Smoothness    =', rss, rsr
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine SPLerr(rss, ip)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), Y(maxN), F(maxN), B(maxN)
      common /BaseI/   nB(maxN, 4, 4)
      common /BaseD/   Bx(maxN, 4), By(maxN, 4)
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
      common /Corect/  H(maxN, maxP)
      common /ABICfn/  SPLdsn, GRDdsn, PNTdsn, Form
      common /Limit/   Xmin, Ymin, Fmin, Xmax, Ymax, Fmax, Gx, Gy
c
      character SPLdsn*70, GRDdsn*70, PNTdsn*70, Form*70
c -------------------------------------------------------------------- c
      write(6, *) ' SPLerr start '
c
      rss = 0.0d0
      do 10 k = 1, nDAT
        if (N(k) .le. 0) go to 10
        totalS = 0.0d0
        do 20 i = 1, 4
          do 20 j = 1, 4
            totalS = totalS + S(nB(k, i, j)) * Bx(k, 5-i) * By(k, 5-j)
20        continue
        B(k)  = F(k)
        do 30 i = 1, nPRM
          B(k)  = B(k) - H(k, i) * S(nSPL+i)
30        continue
        if (ip .eq. 1) then
          Xw = X(k) * (Xmax - Xmin) + Xmin
          Yw = Y(k) * (Ymax - Ymin) + Ymin
          write(13, '(i5, 2f12.3, 3f9.2)')
     &      N(k), Xw, Yw, totalS, B(k), B(k) - totalS
          end if
        rs  = B(k) - totalS
        rss = rss + rs**2
10      continue
      sdx = dsqrt(rss / dfloat(nDAT))
      write(6, '(a, f10.5)') ' Fitting standard deviation = ', sdx
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine SPLwtr(ifile, sd, rss, w1, w2, wABIC)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
c
      common /DataI/   N(maxN), nDAT
      common /Limit/   Xmin, Ymin, Fmin, Xmax, Ymax, Fmax, Gx, Gy
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
c -------------------------------------------------------------------- c
      write(6, *) ' SPLwtr start '
c
      write(ifile, '(3i5)') Mx, My, nPRM
      write(ifile, '(6D13.6)') sd, w1, w2, 0.0d0, 0.0d0, wABIC
      write(ifile, '(4f15.4)') xmin, ymin, xmax, ymax
      write(ifile, '(6D13.6)') (S(i), i=1, nSPL), (errS(i), i=1, nSPL)
      write(ifile, '(a20, 2D15.8)') '       rss,  sd =   ',
     &            rss, dsqrt(rss / dfloat(nDAT))
      write(ifile, '(5h PRM ,i5, 2D13.6)')
     &       (i, S(i+nSPL), errS(i+nSPL), i=1, nPRM)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine GRDwtr(ifile, sd, rss, w1, w2, wABIC)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (maxNx=2000, maxNy=2000)
c
      common /DataI/   N(maxN), nDAT
      common /Limit/   Xmin, Ymin, Fmin, Xmax, Ymax, Fmax, Gx, Gy
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
      common /ABICcm/  ABIC, w1min, w2min
c
      dimension G(maxNx, maxNy), Bx2(4), By2(4)
      character form*8
c -------------------------------------------------------------------- c
      write(6, *) ' GRDwtr start '
c
      Nx = (Xmax - Xmin) / Gx + 1
      Ny = (Ymax - Ymin) / Gy + 1
      if (Nx.gt.maxNx .or. Ny.gt.maxNy) then
        write(6, *) 'Nx.gt.maxNx .or. Ny.gt.maxNy'
        stop
        end if
c
      do 10 ix = 1 , Nx
        do 10 iy = 1 , Ny
          G(ix, iy) = 99999.0
10      continue
c
      Gmin = 9.0d29
      Gmax = -9.0d29
      do 20 ix = 1, Nx
        xw = dfloat(ix-1) * Gx / (Xmax - Xmin)
        if (xw .gt. 1.0d0) go to 20
        rx = xw / Dx
        lx = int(rx) + 1
        if (xw .eq. 1.0d0) lx = Mx
        rx = rx - dfloat(lx - 1)
        call Bsp3(rx, Bx2)
        do 30 iy = 1, Ny
          yw = dfloat(iy-1) * Gy / (Ymax - Ymin)
          if (yw .gt. 1.0d0) go to 30
          ry = yw / Dy
          ly = int(ry) + 1
          if (yw .eq. 1.0d0) ly = My
          ry = ry - dfloat(ly - 1)
          call Bsp3(ry, By2)
          G(ix, iy) = 0.0d0
          do 40 jx = 1, 4
            do 40 jy = 1, 4
              G(ix, iy) = G(ix, iy) +
     &          S(ndim(lx+jx-1, ly+jy-1, Mx)) * Bx2(5-jx) * By2(5-jy)
40          continue
          if (G(ix, iy).lt.Gmin) Gmin = G(ix, iy)
          if (G(ix, iy).gt.Gmax) Gmax = G(ix, iy)
30        continue
20      continue
c
      gw1 = Gmin
      if (Gmin.lt.0.0 .and. Gmin.gt.-10.0) gw1 = 99.0
      if (gw1.lt.0.0) gw1 = -gw1 * 10.0
      gw2 = Gmax
      if (Gmax.lt.0.0 .and. Gmax.gt.-10.0) gw2 = 99.0
      if (gw2.lt.0.0) gw2 = -gw2 * 10.0
      gw = dmax1(gw1, gw2)
c
      if (gw.lt.10.0) then
        form = '(10f8.5)'
        Und = 9.99999
      else if (gw.lt.100.0) then
        form = '(10f8.4)'
        Und = 99.9999
      else if (gw.lt.1000.0) then
        form = '(10f8.3)'
        Und = 999.999
      else if (gw.lt.10000.0) then
        form = '(10f8.2)'
        Und = 9999.99
      else if (gw.lt.100000.0) then
        form = '(10f8.1)'
        Und = 99999.9
      else if (gw.lt.1000000.0) then
        form = '(10f8.0)'
        Und = 999999.
      else
        write(6, *) ' data value over 1000000.0 stop '
        stop
        end if
c
      do 50 ix = 1 , Nx
        do 50 iy = 1 , Ny
          if (G(ix, iy).eq.99999.0) G(ix, iy) = Und
50      continue
c
      write(ifile, '(a14, 6e11.4)') 'sd,w1-w4,abic=',
     &                       sd, w1, w2, 0.0d0, 0.0d0, ABIC
      write(ifile, '(2i5, 2f10.1, 2f10.3, '//form//')')
     &         Nx, Ny, Xmin, Ymin, Gx, Gy, Und, Gmin, Gmax
      write(ifile, form) ((G(ix, iy), iy=1, Ny), ix=1, Nx)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine BxBy
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), Y(maxN), F(maxN), B(maxN)
      common /BaseI/   nB(maxN, 4, 4)
      common /BaseD/   Bx(maxN, 4), By(maxN, 4)
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
c
      dimension Bx2(4), By2(4)
c -------------------------------------------------------------------- c
      do 10 i = 1, nDAT
        if (N(i) .le. 0) go to 10
        rx = X(i) / Dx
        ry = Y(i) / Dy
        nx = int(rx) + 1
        ny = int(ry) + 1
        if (X(i) .ge. 1.0d0) nx = Mx
        if (Y(i) .ge. 1.0d0) ny = My
        rx = rx - dfloat(nx - 1)
        ry = ry - dfloat(ny - 1)
        if (rx.lt.0.0d0 .or. rx.gt.1.0d0) write(6, *) 'BxBy Error'
        if (ry.lt.0.0d0 .or. ry.gt.1.0d0) write(6, *) 'BxBy Error'
        do 20 ix = 1, 4
          do 20 iy = 1, 4
            nB(i, ix, iy) = ndim(nx+ix-1, ny+iy-1, Mx)
20        continue
        call Bsp3(rx, Bx2)
        call Bsp3(ry, By2)
        do 30 j = 1, 4
          Bx(i, j) = Bx2(j)
          By(i, j) = By2(j)
30        continue
10      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      function ndim(ix, iy, Mx)
c -------------------------------------------------------------------- c
      ndim = ix + (iy - 1) * (Mx + 3)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine normal(irc)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), Y(maxN), F(maxN), B(maxN)
      common /Limit/   Xmin, Ymin, Fmin, Xmax, Ymax, Fmax, Gx, Gy
c -------------------------------------------------------------------- c
      irc = 0
      if (nDAT .lt. 20) then
        irc = 999
        write(6, *) '#### Number of points < 20 ####', nDAT
        return
        end if
      do 10 i = 1, nDAT
        if (N(i) .gt. 0) then
          X(i) = (X(i) - Xmin) / (Xmax - Xmin)
          Y(i) = (Y(i) - Ymin) / (Ymax - Ymin)
          end if
10      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
c   Modified Cholesky compact version   U(i,i) = 1                     c
c              use U2 dimenstion                                       c
c -------------------------------------------------------------------- c
      subroutine solveZ(S, nSPL, nPRM, iback)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common /matrix/ U(nRcmp, nCcmp), nROW
c
      dimension S(maxM), w(maxM)
c -------------------------------------------------------------------- c
      write(6, *) ' solveZ start (Compact modified Cholesky)'
c
      m = nSPL + nPRM
      do 10 k = 1, nSPL
        nstr = max(   1, k-nROW+1)
        nfin = min(nSPL, k+nROW-1)
        rU = U(1, k)
        do 20 i = nstr, k-1
          kri = k - i + 1
          w(i) = U(kri, i) * U(1, i)
          rU = rU - w(i) * U(kri, i)
20        continue
        U(1, k) = rU
c       if (rU .lt. 0.0) U(1, k) = -rU
        do 30 j = k+1, nfin
          jrk = j - k + 1
          rU = U(jrk, k)
          do 40 i = max(1, j-nROW+1), k-1
            rU = rU - w(i) * U(j-i+1, i)
40          continue
          U(jrk, k) = rU / U(1, k)
30        continue
        if (nPRM .ne. 0) then
          do 31 j = nSPL+1, m
            jr = nROW + j - nSPL
            rU = U(jr, k)
            do 41 i = nstr, k-1
              rU = rU - w(i) * U(jr, i)
41            continue
            U(jr, k) = rU / U(1, k)
31          continue
          end if
10      continue
c
      if (nPRM .ne. 0) then
        do 12 k = nSPL+1, m
          kr = nROW + k - nSPL
          rU = U(kr, k)
          do 22 i = 1, k-1
            w(i) = U(kr, i) * U(1, i)
            rU = rU - w(i) * U(kr, i)
22          continue
          U(1, k) = rU
          do 32 j = k+1, m
            jr = nROW + j - nSPL
            rU = U(jr, k)
            do 42 i = 1, k-1
              rU = rU - w(i) * U(jr, i)
42            continue
            U(jr, k) = rU / U(1, k)
32          continue
12        continue
        end if
c
      if (iback .eq. 1) then
        do 50 k = 1, nSPL
          rU = U(nROW+nPRM+1, k)
          do 60 i = max(1, k-nROW+1), k-1
            rU = rU - U(k-i+1, i) * S(i)
60          continue
          S(k) = rU
50        continue
        if (nPRM .ne. 0) then
          do 51 k = nSPL+1, m
            kr = nROW + k - nSPL
            rU = U(nROW+nPRM+1, k)
            do 61 i = 1, k-1
              rU = rU - U(kr, i) * S(i)
61            continue
            S(k) = rU
51          continue
          end if
        if (nPRM .ne. 0) then
          do 71 k = m, nSPL+1, -1
            rU = S(k) / U(1, k)
            do 81 i = k+1, m
              rU = rU - U(nROW+i-nSPL, k) * S(i)
81            continue
            S(k) = rU
71          continue
          end if
        do 70 k = nSPL, 1, -1
          rU = S(k) / U(1, k)
          do 80 i = k+1, min(nSPL, nROW+k-1)
            rU = rU - U(i-k+1, k) * S(i)
80          continue
          if (nPRM .ne. 0) then
            do 82 i = nSPL+1, m
              rU = rU - U(nROW+i-nSPL, k) * S(i)
82            continue
            end if
          S(k) = rU
70        continue
        end if
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine lndetU(m, detU)
c -------------------------------------------------------------------- c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common /matrix/ U(nRcmp, nCcmp), nROW
c -------------------------------------------------------------------- c
      write(6, *) ' lndetU start '
c
      detU = 0.0d0
      do 10 nC = 1, m
        if (U(1, nC) .gt. 0.0d0) then
          detU = detU + dlog(U(1, nC))
        else
          detU = detU + dlog(-U(1, nC))
          write(6, *) ' lndetU U(1, nC) < 0.0', nC , m, U(1, nC)
          end if
10      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine retrnE(ePRM, nC)
c -------------------------------------------------------------------- c
      implicit    real*8 (a-h, o-z)
      parameter   (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter   (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter   (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common     /matrix/ U(nRcmp, nCcmp), nROW
c -------------------------------------------------------------------- c
      write(6, *) ' retrnE start '
c
      ePRM = dsqrt(1.0d0 / U(1, nC))
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine errorS(errS, nSPL, nPRM)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common /matrix/ U(nRcmp, nCcmp), nROW
c
      dimension errS(maxM), Z(maxM)
c -------------------------------------------------------------------- c
      write(6, *) ' errorS start '
c
      m = nSPL + nPRM
      do 10 k = 1, nSPL
        Z(k) = 1.0d0
        do 20 j = k+1, nSPL
          rU = 0.0d0
          do 30 i = max(k, j-nROW+1), j-1
            rU = rU - U(j-i+1, i) * Z(i)
30          continue
          Z(j) = rU
20        continue
        if (nPRM .ne. 0) then
          do 21 j = nSPL+1, m
            jr = nROW + j - nSPL
            rU = 0.0d0
            do 31 i = k, j-1
              rU = rU - U(jr, i) * Z(i)
31            continue
            Z(j) = rU
21          continue
          end if
        if (nPRM .ne. 0) then
          do 41 j = m, nSPL+1, -1
            rU = Z(j) / U(1, j)
            do 51 i = j+1, m
              ir = nROW + i - nSPL
              rU = rU - U(ir, j) * Z(i)
51            continue
            Z(j) = rU
41          continue
          end if
        do 40 j = nSPL, k, -1
          rU = Z(j) / U(1, j)
          do 50 i = j+1, min(nSPL, j+nROW-1)
            rU = rU - U(i-j+1, j) * Z(i)
50          continue
          if (nPRM .ne. 0) then
            do 52 i = nSPL+1, m
              ir = nROW + i - nSPL
              rU = rU - U(ir, j) * Z(i)
52            continue
            end if
          Z(j) = rU
40        continue
        errS(k) = Z(k)
10      continue
c
      do 13 k = nSPL+1, m
        Z(k) = 1.0d0
        do 23 j = k+1, m
          jr = nROW + j - nSPL
          rU = 0.0d0
          do 33 i = k, j-1
            rU = rU - U(jr, i) * Z(i)
33          continue
          Z(j) = rU
23        continue
        do 43 j = m, k, -1
          rU = Z(j) / U(1, j)
          do 53 i = j+1, m
            ir = nROW + i - nSPL
            rU = rU - U(ir, j) * Z(i)
53          continue
          Z(j) = rU
43        continue
        errS(k) = Z(k)
13      continue
      do 60 k = 1, m
        errS(k) = dabs(errS(k))
60      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine obserZ
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common /DataI/   N(maxN), nDAT
      common /DataD/   X(maxN), Y(maxN), F(maxN), B(maxN)
      common /BaseI/   nB(maxN, 4, 4)
      common /BaseD/   Bx(maxN, 4), By(maxN, 4)
      common /SplineI/ nSPL, nPRM, Mx, My
      common /Corect/  H(maxN, maxP)
      common /matrix/  U(nRcmp, nCcmp), nROW
      common /matrix2/ U2(nRcmp, nCcmp), nU2
c
      dimension  me(4)
      data       me/4*1/
c -------------------------------------------------------------------- c
      write(6, *) ' obserZ start '
c
      nROW = Mx * 3 + 13
      if (nU2 .eq. 999) go to 1
c
      do 10 i = 1, nDAT
        do 20 j = 1, nPRM
          k = nSPL + j
          U(nROW+nPRM+1, k) = U(nROW+nPRM+1, k) + H(i, j) * F(i)
          do 20 l = j, nPRM
            m = nROW + l
            U(m, k) = U(m, k) + H(i, j) * H(i, l)
20        continue
        if (N(i) .le. 0) go to 10
        do 30 ix1 = 1, 4
          do 30 iy1 = 1, 4
            n1 = nB(i, ix1, iy1)
            p1 = Bx(i, 5-ix1) * By(i, 5-iy1)
            do 40 j = 1, nPRM
              U(nROW+j, n1) = U(nROW+j, n1) + p1 * H(i, j)
40            continue
            U(nROW+nPRM+1, n1) = U(nROW+nPRM+1, n1) + p1 * F(i)
            me(iy1) = ix1
            do 50 ix2 = 4, iy1, -1
              do 50 iy2 = 4, me(ix2), -1
                n2 = nB(i, iy2, ix2)
                n3 = n2 - n1 + 1
                p2 = Bx(i, 5-iy2) * By(i, 5-ix2)
                U(n3, n1) = U(n3, n1) + p1 * p2
50            continue
            me(iy1) = 1
30        continue
10      continue
c
      do 60 i = 1, nCcmp
        do 60 j = 1, nRcmp
          U2(j, i) = U(j, i)
60      continue
      nU2 = 999
      return
c
1     continue
      do 70 i = 1, nCcmp
        do 70 j = 1, nRcmp
          U(j, i) = U2(j, i)
70      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine clearZ
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common /matrix/ U(nRcmp, nCcmp), nROW
c -------------------------------------------------------------------- c
      write(6, *) ' clearZ start '
c
      do 10 nC = 1, nCcmp
        do 10 nR = 1, nRcmp
          U(nR, nC) = 0.0d0
10      continue
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine SMTerr(w1, w2, rsr, detD)
c
      implicit    real*8 (a-h, o-z)
      parameter   (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter   (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter   (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
      common /matrix/  U(nRcmp, nCcmp), nROW
c -------------------------------------------------------------------- c
      write(6, *) ' SMTerr start '
c
      call clearZ
      call priorZ(w1, w2, rsr, 1)
      call solveZ(S, nSPL, 0, 0)
      call lndetU(nSPL-1, detD)
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine priorZ(w1, w2, rsr, ip)
c
      implicit  real*8 (a-h, o-z)
      parameter (maxN=15000, maxMx=100, maxMy=100, maxP=3120)
      parameter (maxM=(maxMx+3)*(maxMy+3)+maxP)
      parameter (nRcmp=15+maxMx*3+maxP-1, nCcmp=maxM+1)
c
      common /SplineI/ nSPL, nPRM, Mx, My
      common /SplineD/ S(maxM), errS(maxM), Dx, Dy
      common /matrix/  U(nRcmp, nCcmp), nROW
c
      dimension me(4), bi(3, 4, 4)
c -------------------------------------------------------------------- c
      write(6, '(a, 2d15.6)') ' priorZ start ', w1, w2
c
      w10 = w1 * Dy    /  Dx
      w01 = w1 * Dx    /  Dy
      w20 = w2 * Dy    /  Dx**3
      w11 = w2 * 2.0d0 / (Dx*Dy)
      w02 = w2 * Dx    /  Dy**3
c
      call intB(bi)
c
      if (ip .eq. 1) rsr = 0.0d0
c
      do 10 iw = 1, Mx
        do 10 jw = 1, My
          do 10 lw = 1, 4
            do 10 kw = 1, 4
              n1 = ndim(iw+kw-1, jw+lw-1, Mx)
              me(4) = 1
              me(3) = 1
              me(2) = 1
              me(lw) = kw
              do 10 nw = 4, lw, -1
                p1 = w10 * bi(1, lw, nw)
                p2 = w01 * bi(2, lw, nw)
                p3 = w20 * bi(1, lw, nw)
                p4 = w11 * bi(2, lw, nw)
                p5 = w02 * bi(3, lw, nw)
                n2 = ndim(iw+3, jw+nw-1, Mx)
                do 10 mw = 4, me(nw), -1
                  w = p1 * bi(2, kw, mw) + p2 * bi(1, kw, mw) +
     &                p3 * bi(3, kw, mw) + p4 * bi(2, kw, mw) +
     &                p5 * bi(1, kw, mw)
                  n3 = n2 - n1 + 1
                  U(n3, n1) = U(n3, n1) + w
                  if (ip .eq. 1) then
                    if (n1 .ne. n2) w = w * 2.0d0
                    rsr = rsr + S(n1) * S(n2) * w
                    end if
                  n2 = n2 - 1
10      continue
      if (ip .eq. 1 .and. rsr .lt. 0.0d0) then
        write(6, *) 'rsr < 0.0', rsr
c       rsr = 0.0d0
        rsr = -rsr
        end if
c -------------------------------------------------------------------- c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine Bsp3(r, B)
c
      implicit real*8 (a-h, o-z)
      dimension B(4)
c -------------------------------------------------------------------- c
      r2 = r * r
      r3 = r * r * r
c
      a = 1.0d0 / 6.0d0
      B(1) = r3 * a
      B(2) = (3.0d0 * (-r3 + r2 + r) + 1.0d0) * a
      B(3) = (3.0d0 * r3 - 6.0d0 * r2 + 4.0d0) * a
      B(4) = (-r3 + 3.0d0 * (r2 - r) + 1.0d0) * a
c
      return
      end
c -------------------------------------------------------------------- c
c -------------------------------------------------------------------- c
      subroutine intB(bi)
c
      implicit  real*8 (a-h, o-z)
c
      dimension bi(3,4,4), a(3,4), b(3,4), c(3,4), d(3,4), fac(3)
c
      data a   / 1.0d0, 0.0d0, 0.0d0,-3.0d0, 0.0d0, 0.0d0,
     &           3.0d0, 0.0d0, 0.0d0,-1.0d0, 0.0d0, 0.0d0/
      data b   / 0.0d0, 1.0d0, 0.0d0, 3.0d0,-3.0d0, 0.0d0,
     &          -6.0d0, 3.0d0, 0.0d0, 3.0d0,-1.0d0, 0.0d0/
      data c   / 0.0d0, 0.0d0, 1.0d0, 3.0d0, 2.0d0,-3.0d0,
     &           0.0d0,-4.0d0, 3.0d0,-3.0d0, 2.0d0,-1.0d0/
      data d   / 0.0d0, 0.0d0, 0.0d0, 1.0d0, 1.0d0, 1.0d0,
     &           4.0d0, 0.0d0,-2.0d0, 1.0d0,-1.0d0, 1.0d0/
      data fac /36.0d0, 4.0d0, 1.0d0/
c -------------------------------------------------------------------- c
      do 10 j = 1, 3
      do 10 k = 1, 4
      do 10 l = 1, 4
c
       bi(j,5-k,5-l) =
     1   (a(j,k)*a(j,l)                                          )/7.0d0
     2  +(a(j,k)*b(j,l)+b(j,k)*a(j,l)                            )/6.0d0
     3  +(a(j,k)*c(j,l)+b(j,k)*b(j,l)+c(j,k)*a(j,l)              )/5.0d0
     4  +(a(j,k)*d(j,l)+b(j,k)*c(j,l)+c(j,k)*b(j,l)+d(j,k)*a(j,l))/4.0d0
     5  +(              b(j,k)*d(j,l)+c(j,k)*c(j,l)+d(j,k)*b(j,l))/3.0d0
     6  +(                            c(j,k)*d(j,l)+d(j,k)*c(j,l))/2.0d0
     7  +(                                          d(j,k)*d(j,l))/1.0d0
c
        bi(j,5-k,5-l) = bi(j,5-k,5-l) / fac(j)
10    continue
c
      return
      end
